/*
common.h; for use with whitespd.c and the associated subroutines.
(c)2023, 2024 Jonathon David White (jonathondavid@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see 
<https://www.gnu.org/licenses/gpl-3.0.en.html>.

If you choose to make use of this program or subroutines in your work, 
we would appreciate it if you would cite our paper 
"Building Circadian Effective Spectra: A C Language Toolkit"

You may download the latest version of the code from https://www.whiteslight.com
*/



#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#define FLG_ERR2SCREEN 0  // Send error messages to screen (0:No, 1:YES) (USER MAY CHANGE)
#define MCH 8          // maximum number of different LED types (Channels)  (USER MAY CHANGE)
#define MINTERVAL 256  // maximum number of set point SPDs for a 24 hour period (USER MAY CHANGE)

#define H 6.626e-34    // Plank's Contant  
#define C 2.99e8       // Speed of light
#define NM2M 1E-9      // conversion factor:  m/nm
#define EPS 0.0000001    // Small Number 
#define DUV_LIMIT_CRI 0.08  // if duv>limit, no CRI    0.0054 

#define NM_START 380   // starting wavelength for visible SPD
#define NM_END 831     // ending wavelength for visible SPD 
#define MFNAME 128     // maximum length of a file name.
#define MCOND 3        // maximum conditions to return optimized configuration (e.g. score, mEDI, Lux)
#define MB 512        // maximum buffer size
#define MTCS 15 

/* array subscripts */
#define X 0
#define Y 1
#define Z 2
#define DUV_ACTUAL 0
#define DUV_TARGET 1
#define DUV_DELTA 2
#define COST_FRONT 0   // Up front cost of the LED strips
#define POWER 1        // Instanteous Power consumption of the LED strips

char bb[MB];
char bb2[MB];

struct SPD_point{ double nm; double y;}; // SPD raw data point, Intensity(wavelength)
struct SPD_weight{ char fname[MFNAME]; double fraction[MCH]; char comment[MB]; double score;};  
struct Coord3{ double w; double u; double v;};
struct Coord{ double x; double y; };

struct SPD{ // complete spd: energy/unit area/time/nm
	char fname[MFNAME]; // name of SPD file 
	double y[NM_END+1];  // SPD
	double cost[2]; 	// upfront US$, power consumption [W/unit]
	char stripInfo[MB]; // supplier, manufacture date, leds/m ...
	double tc_ms;       // time constant data take with
};

#define MALPHA 6
#define _R 3
#define _M 4
#define _PLUX 5


struct SPDparameters{
	double cie[3], ciesum, ciexy[2]; // cie X,Y,Z & SUM
	double duv[3];    // ANSI C78.377-2017	; Duv calculated, target, offset
	double cct;       // CCT [K]
	double cri[2];   // cri8 and cri14
	double a_opic[MALPHA]; // S-, M-, L-cone, Rhodopic, Melanopic ELR + Photopic illuminance (horizontal) plane
	double mEDI;     // m-EDI lx (vertical plane)
	double irrad;    // irradiance
	double cost[2];  // upfront US$, power consumption [W/unit]
	double user[2];  // comfort, usefulness [0 5]
	double mEDIperW; // m-EDI Lux/W    
	double lxperW;   // photopic lx/W (horizontal plane) 
	double score;    // Negative means does not meet minimum requirements 
};
struct LightConfig{
	double etaG; // geometry factor = Horizontal/Vertical Lux 
	double refEh; // single measurement of photopic illuminance on horizontal plane, 100% power, one strip 8
	double simEh; // amount returned by simulation using an the experimental factor
	char s; // source of the scaling Lux.  'b' is Baole 21 paper conversion (default), '-' deduce from current input data
	char spd[MB]; // reference SPD filename
	double scaleFac; //  (measured photopic lx)/ (value obtained in simulation) 
};

// Represent how to score, specify goals/criteria in brute force, permutation and ai modes.
#define LIMITMAX 10    // reject if > criteria (e.g. |Duv| > 0.006)
#define LIMITMIN -10   // reject if < criteria (e.g. CRI-Ra < 80)
#define MAXIMIZE 1     // GOAL: maximize parameter (e.g. lx)
#define MINIMIZE -1    // GOAL: minimize parameter (e.g. Power Consumption, Duv)
#define DONTCARE 0     // Not Important Parameter
#define REJECT -10000  // This value to indicates SPD outside limits 
struct Goal{
	int duv;
	int cct;
	int cri[2];
	int a_opic[MALPHA];
	int mEDI;
	int irrad;
	int cost[2];
	int user[2];
	int mEDIperW;
	int lxperW;
	struct SPDparameters p;	
};



/* GLOBAL OPTICAL CONSTANTS */
struct SPD ciebar[3]; 		// CIE1931 (constants)
struct SPD action[MALPHA];	// a-optic (constants)
struct SPD cri_s[3];   	// CRI (constants)
struct SPD tcs[MTCS]; 		// CRI (constants)

/* ******************** SUBROUTINES ************************** */
/* in whitespd.c -- called by main() */
void whiteSPD(char *fname_weight,char *mode,int flgWriteSPD);
/* buildspd.c */
struct SPD buildSPD(struct SPD_weight w,struct SPD spdBasis[]);
/* score.c */
double score(struct SPDparameters p,struct SPDparameters goal, struct SPDparameters w);
/* getWeight.c */
struct SPD_weight readWeight(FILE *fweight);
struct SPD_weight getWeight(FILE *fweight,struct SPD_weight wprev,int nch,char *mode);

/* file cri.c */
struct SPDparameters calcCRI(struct SPD s, struct SPD bar[], struct SPD cri_s[],
	struct SPD tcs[], struct SPDparameters p); 
/* tradparam.c */
struct SPDparameters calcCIE1931(struct SPD s, struct SPD bar[], struct SPDparameters p);
double calcCCT(struct SPDparameters p);
struct Coord calcUV(struct SPDparameters p);
double calcDUV(struct Coord uv);
double calcDUVtarget(double t);
double calcIrrad(struct SPD s);
/* alpha.c */
struct SPDparameters calcAlphaOpicELR(struct SPD s, struct SPD action[], struct SPDparameters p);
/* file calcpar.c */
struct SPDparameters calcSPDpar(struct SPD s,struct LightConfig e, struct Goal g);
/* file gnuplotSPD.c : Plotting of data */
int gnuplotSPDopen(char *fname_plt, char *title,char *ylabel); 
int gnuplotSPD(char *dirout, char *fname,char *fname_plot,char type); 
int gnuplotSPDclose(char *fname_plt,int flgPause);                   
void plotSPD(char fname_weight[]);
/* in read.c */
struct SPD readSPD(char *fname);
struct SPD readInterpSPD(char *fname);
struct LightConfig readLightConfig(void);
void readOpticalConstants(void); 	//used to read all the OPTICS constants
struct Goal readGoal(void); //read optimization goal

/*file write.c*/
void writeSPDusage(void);
int writeSPD(struct SPD spd,int nm_step, int flgCol, char *dirout); 
void writeSPDparHeader(FILE *fo, char *line);
int writeSPDpar(FILE *fo,int iline,char *fname,struct SPDparameters p,struct SPD_weight w,int nch);
struct SPDparameters readSPDpar(FILE *f,struct SPDparameters p,struct SPD_weight *w,int nch);
/* file: prnlog.c */
#define PRNLOG_OFF 0
#define PRN2FILE 1                 
#define PRN2ALL 2
#define PRN2ERR 3
#define PRNLOG PRN2ALL
char bprn[MB];
void prnlog(char *b, char flgPrn); 

