/*
whitespd.c; builds, calculates optical parameters, and optimizes SPD for LED strip lighting 
(c)2023, 2024 Jonathon David White (jonathondavid@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see 
<https://www.gnu.org/licenses/gpl-3.0.en.html>.

If you choose to make use of this program or subroutines in your work, 
we would appreciate it if you would cite our paper 
"Building Circadian Effective Spectra: A C Language Toolkit"

You may download the latest version of the code from https://www.whiteslight.com
*/

#include "common.h"
#define STATUS_UPDATE_FREQUENCY 1000000
#define VS_WHITESPD 240516
void whiteSPD(char *fname_weight,char *mode,int flgWriteSPD){
	int ich, nch,iwrite=0,k;
	long int iline=0;
	char namesBasisSPDs[MB]; //number and file names of input SPDs
	FILE *fo, *fweight;
	time_t start = time(NULL), end;	
	struct SPD_weight w={"n/a",{0,0,0} ,"\0",-1000.0}; // weight of input SPDs
	struct SPD_weight wbest[MCOND]={{"best",{0,0,0} ,"\0",-1000.0}}; //multi-param optimizing
	struct SPD spdBasis[MCH]; // input SPDs (to combine or obtain parameters)
	struct SPD s; 	// SPD for which to obtain parameters	
	struct SPDparameters p; // parameters of currently being evaluated SPD
	struct SPDparameters pbest[MCOND] ={{ {0,0,0},0,{0,0},{0,0,0},0,{0,0},{0,0,0,0,0,0},
		0,0,{1e6,1e6},{0,0},0,0,REJECT }}; //multi-param max
	struct Goal goal=readGoal(); // Target goal(s) for scoring.
	struct LightConfig e=readLightConfig();  // including measured ratio of illuminance on horizontal:vertical plane
	readOpticalConstants(); // all these are global variables	

	if((fweight=fopen(fname_weight,"r"))==NULL)goto errOpenWeight;
	if((fo=fopen("out/spdpar.csv","w"))==NULL)goto errOpenOut;	
	fgets(namesBasisSPDs,MB,fweight);
	writeSPDparHeader(fo,namesBasisSPDs);	
	nch=atol(strtok(namesBasisSPDs,",")); // first line of weight file gives number + names of input SPDs
	fprintf(stderr,"\nwhiteSPD(%d) reading basis SPDs from %s  nch=%d",VS_WHITESPD,fname_weight,nch);
	for(ich=0;ich<nch;ich++)spdBasis[ich]=readSPD(strtok(NULL,","));
	fprintf(stdout,"\nwhiteSPD(%d) Setup complete.  Starting parameter analysis...\n",VS_WHITESPD);
	w=getWeight(fweight,w,nch,mode);
	while(w.fraction[0]>-0.1){
		iline++;
		s=buildSPD(w,spdBasis);
		if(flgWriteSPD)writeSPD(s,1,2,"out");		
		p=calcSPDpar(s,e,goal);         
		w.score=p.score;
		switch(mode[0]){
			case 'r': case 'R': writeSPDpar(fo,iline,s.fname,p,w,nch); break;
			case 'b': case 'p': case 'B' : case 'P':
				if(p.score>pbest[0].score){wbest[0]=w; pbest[0]=p;}	/* save spectra best meeting goal (score) */
				if(p.mEDIperW>pbest[1].mEDIperW){pbest[1]=p;wbest[1]=w;} /* save best mEDI/W  */
				if(p.mEDI>pbest[2].mEDI){pbest[2]=p;wbest[2]=w;}   /* save best mEDI */
				if(flgWriteSPD && p.score>(REJECT+1)){writeSPDpar(fo,iline,s.fname,p,w,nch); iwrite++;}
				if(iline%STATUS_UPDATE_FREQUENCY==0){ // status update for running program...
					fprintf(stderr,"\r%ld",iline);
					if(iline%(STATUS_UPDATE_FREQUENCY*10)==0){
						for(k=0;k<MCOND;k++)writeSPDpar(stderr,iline,s.fname,pbest[k],wbest[k],nch);
						writeSPDpar(stderr,iline,"Testing",p,w,nch);
						fprintf(stderr,"\n");
					}
				}
				break;
			default : fprintf(stderr,"whiteSPD(): ERR mode=%s('%c') does not exist."
				"\n\tonly brute, perm and read modes active in this version",mode,mode[0]); exit(0);
		}
		w=getWeight(fweight,w,nch,mode);
	}
					
	if(mode[0]=='b' || mode[0]=='p')
		for(k=0;k<MCOND;k++)writeSPDpar(fo,iline,"optimum",pbest[k],wbest[k],nch);
	end = time(NULL);
	fprintf(stdout,"\nwhiteSPD(%d)\n\texecution time = %.0lfs \n\t%d SPDs processed"
		"\n\t%d SPDs accepted, parameters written",VS_WHITESPD,difftime(end, start),iline,iwrite);
	fclose(fo);
	fclose(fweight);
	return;
errOpenWeight:fprintf(stderr,"\nwhiteSPD() ERR Can't open '%s' for input",fname_weight);exit(4);
errOpenOut: fprintf(stderr,"\nwhiteSPD() ERR Can't open out/spdpar.csv");exit(4);
}
#define VS_MAIN 241002
int main(int argc, char *argv[]){
	int flgWriteSPD;
	sprintf(bprn,"whitespd(%d): Copyright (C)2023,2024 XXX\n"
		"This program comes with ABSOLUTELY NO WARRANTY;\n"
		"This is free software, and you are welcome to redistribute it \n"
		"under certain conditions (see readme.pdf)\n"
		"(Note for CRI, Duv<%.5lf required.)"
		,VS_MAIN,DUV_LIMIT_CRI);
	prnlog(bprn,PRNLOG);
	if(argc!=4)goto err;
	flgWriteSPD=atol(argv[3]);
	sprintf(bprn,"\nreading weights from %s running in %s mode\n",argv[1],argv[2]);
	prnlog(bprn,PRNLOG);
	if(flgWriteSPD && argv[2][0]=='r')sprintf(bprn,"Writing both parameters and built SPDs");
	if(flgWriteSPD && argv[2][0]!='r')sprintf(bprn,"Writing parameters for all acceptable SPDs");
	if(!flgWriteSPD && argv[2][0]!='r')sprintf(bprn,"Finding best SPDs");
	if(!flgWriteSPD && argv[2][0]=='r')sprintf(bprn,"Writing parameters only for SPDs");	
	prnlog(bprn,PRNLOG);
	whiteSPD(argv[1],argv[2],flgWriteSPD);
	if(flgWriteSPD && argv[2][0]=='r')plotSPD(argv[1]);
	
	return(0);
err: fprintf(stderr,"\nwhitespd(%d) USAGE: 'whitespd fname.csv mode flg_saveSPD' (see readme.pdf)",VS_MAIN); 
	return(1);
}